"""A module containing the command-line interface for Directory WAND.

This module contains the Parser class for parsing command line arguments. The
Parser class is a wrapper around the argparse.ArgumentParser class that adds
some standardised arguments and enables arbitrarily many arbitrarily named
arguments.
"""

import argparse


class StoreDictKeyPair(argparse.Action):
    """A class for storing key-value pairs from the command line."""

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Store the key-value pair in the namespace.

        Args:
            parser (argparse.ArgumentParser):
                The parser instance.
            namespace (argparse.Namespace):
                The namespace to store the key-value pair in.
            values (str):
                The key-value pair to store.
            option_string (str):
                The option string.
        """
        key = option_string.lstrip("--")
        if not hasattr(namespace, "replacements"):
            setattr(namespace, "replacements", {})
        getattr(namespace, "replacements")[key] = values


class StoreListKeyPair(argparse.Action):
    """A class for storing key-value pairs from the command line."""

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Store the key-value pair in the namespace.

        Args:
            parser (argparse.ArgumentParser):
                The parser instance.
            namespace (argparse.Namespace):
                The namespace to store the key-value pair in.
            values (str):
                The key-value pair to store.
            option_string (str):
                The option string.
        """
        key = option_string.lstrip("-")
        if not hasattr(namespace, "list_args"):
            setattr(namespace, "list_args", {})
        namespace.list_args[key] = values


class Parser(argparse.ArgumentParser):
    """
    A class for parsing command line arguments.

    NOTE: This class is a wrapper around the argparse.ArgumentParser class
    adding some standardised arguments and enabling arbitrarily many
    arbitrarily named arguments.

    Attributes:
        parser (argparse.ArgumentParser):
            The parser instance.
    """

    def __init__(self, description):
        """
        Create the parser instance.

        Args:
            description (str):
                A description of the program.
        """
        # Create the parser
        super(Parser, self).__init__(description=description)

        # Add the template argument
        self.add_argument(
            "template",
            type=str,
            help="The template to use for the model.",
        )

        # Add an optional path to the root for the outputs
        self.add_argument(
            "--root",
            type=str,
            default=".",
            help="The root directory for the outputs.",
        )

        # Add arbitrary arguments
        self.add_argument(
            "--",
            dest="replacements",
            action=StoreDictKeyPair,
            nargs=1,
            metavar="VALUE",
            help="A key-value pair for a placeholder replacement. Should "
            "be in the form --key value, where key is the name ({name}) to "
            "replace in directory paths and files, and value is an "
            "inclusive range (1-5).",
        )
        self.add_argument(
            "-",
            dest="replacements",
            action=StoreListKeyPair,
            nargs="+",
            metavar="VALUE",
            help="A key-value pair for a placeholder replacement. Should "
            "be in the form -key value1 value2 value3, where key is the name "
            "({name}) to replace in directory paths and files, and value is "
            "a list.",
        )

    def __str__(self):
        """Summarise the command line arguments."""
        self.print_help()
        return ""

    def parse_args(self):
        """
        Parse the command line arguments.

        This will handle any "unknown" arguments that are passed to the
        command line by storing them in a dictionary called "replacements"
        attached to the returned namespace.

        Returns:
            argparse.Namespace:
                The parsed command line arguments.
        """
        # Parse arguments
        args, unknown_args = self.parse_known_args()

        args.replacements = {}

        # Process unknown_args manually
        while unknown_args:
            if unknown_args[0].startswith("--"):
                key = unknown_args.pop(0).lstrip("--")
                if (
                    unknown_args
                    and not unknown_args[0].startswith("-")
                    and not unknown_args[0].startswith("--")
                ):
                    value = unknown_args.pop(0)
                    args.replacements[key] = value
                else:
                    args.replacements[key] = None
            elif unknown_args[0].startswith("-"):
                key = unknown_args.pop(0).lstrip("-")
                values = []
                while (
                    unknown_args
                    and not unknown_args[0].startswith("-")
                    and not unknown_args[0].startswith("--")
                ):
                    values.append(unknown_args.pop(0))
                args.replacements[key] = values
            else:
                unknown_args.pop(0)

        return args



"""A module contain the class defining the file.

A File contains all the information to make copies of a template file.

Example:
    # Create a file instance
    file = File("path/to/file")

    # Check if the file is a softlink
    file.is_softlink

    # Check if the file is executable
    file.is_executable

    # Get the placeholders
    file.get_placeholders()

    # Make a copy of the file with the placeholders swapped out
    file.make_copy_with_swaps("path/to/new/file", placeholder="value")

"""

import os
import re


class File:
    """
    A class for defining the file.

    A file contains all the information to make copies of a template file.

    A file can be a softlink, executable, hidden, empty, or a text based file
    containing placeholders.

    Attributes:
        path (str):
            The path to the file.
        name (str):
            The name of the file.
        is_softlink (bool):
            Whether the file is a softlink.
        is_executable (bool):
            Whether the file is executable.
        is_hidden (bool):
            Whether the file is hidden.
        is_empty (bool):
            Whether the file is empty.
        is_text (bool):
            Whether the file is text.
        has_placeholders (bool):
            Whether the file has placeholders.
    """

    def __init__(self, path):
        """
        Create the file instance.

        As well creating the File instance, we will also extract any
        placeholders and their line index for quicker swapping later.

        Args:
            path (str):
                The path to the file.
        """
        # Path to the file
        self.path = path

        # If this file contains any placeholders we'll store there line index
        # for quicker replacing later
        self._place_holder_lines = []
        self._placeholders = set()
        self.get_placeholders()

    def __str__(self):
        """
        Return the string representation of the file.

        Returns:
            str:
                The string representation of the file.
        """
        return self.path

    @property
    def name(self):
        """Return the name of the file."""
        return os.path.basename(self.path)

    @property
    def is_softlink(self):
        """Return whether the file is a softlink."""
        return os.path.islink(self.path)

    @property
    def is_executable(self):
        """Return whether the file is executable."""
        return os.access(self.path, os.X_OK)

    @property
    def is_hidden(self):
        """Return whether the file is hidden."""
        return self.path.startswith(".")

    @property
    def is_empty(self):
        """Return whether the file is empty."""
        return os.stat(self.path).st_size == 0

    @property
    def is_text(self):
        """Return whether the file is text."""
        try:
            with open(self.path, "r") as file:
                file.read()
            return True
        except UnicodeDecodeError:
            return False

    @property
    def has_placeholders(self):
        """Return whether the file has placeholders."""
        return len(self._placeholders) > 0

    def get_placeholders(self):
        """Extract what place holders a file contains."""
        # If the file isn't text we can't extract placeholders
        if not self.is_text:
            return

        # Open the file and read the lines
        with open(self.path, "r") as file:
            lines = file.readlines()

        # Define the regex pattern to extract the placeholders
        # (e.g. {placeholder})
        pattern = re.compile(r"\{([a-zA-Z0-9_]+)\}")

        # Iterate over the lines and extract the placeholders
        for index, line in enumerate(lines):
            # Find all the matches in the line
            matches = pattern.findall(line)

            # If we have matches...
            if matches is not None:
                # Store the line index for later
                self._place_holder_lines.append(index)

                # Add the placeholders to the set
                for match in matches:
                    self._placeholders.add(match)

    def make_copy_with_swaps(self, path, **swaps):
        """
        Make a copy of the file with the placeholders swapped out.

        This will correctly handle softlinks, executable files, and files
        with placeholders. If the file has placeholders, you must provide
        the values to swap out, otherwise an error will be raised.

        Args:
            path (str):
                The path to save the new file.
            **swaps:
                The placeholders to swap out.
        """
        # Get the new file path handling any possible placeholders
        path += "/" if not path.endswith("/") else ""
        path = path + self.name.format(**swaps)

        # Before checking for swaps we might have a softlink, so we need to
        # copy the link
        if self.is_softlink:
            link = os.readlink(self.path)
            os.symlink(link, path)
            return

        # We might have an executable file, so we need to copy the permissions
        # and then the file
        if self.is_executable:
            # Copy the file
            with open(self.path, "rb") as file:
                with open(path, "wb") as new_file:
                    new_file.write(file.read())

            # Copy the permissions
            os.chmod(path, os.stat(path).st_mode)
            return

        # We might only need to make a straight copy
        if not self.has_placeholders:
            # Copy the file
            with open(self.path, "rb") as file:
                with open(path, "wb") as new_file:
                    new_file.write(file.read())
            return

        # OK, we dont have a simple case, lets handle the placeholders

        # Ensure we have all the placeholders before we start swapping
        missing = self._placeholders - set(swaps.keys())
        if len(missing) > 0:
            raise ValueError(f"Missing placeholders: {missing}")

        # Open the file and read the lines
        with open(self.path, "r") as file:
            lines = file.readlines()

        # Iterate over the lines and swap out the placeholders
        for index in self._place_holder_lines:
            # Get the line
            line = lines[index]

            line = line.format(**swaps)

            # Update the line
            lines[index] = line

        # Write the new file
        with open(path, "w") as file:
            file.writelines(lines)



"""A module defining the directory class.

A Directory contains all the information to make copies of a template
directory. A directory is a tree structure containing files and other
directories.

Example:
    # Create a directory instance
    directory = Directory("path/to/directory")

    # Check if the directory is empty
    directory.is_empty

    # Check if the directory contains placeholders
    directory.contains_placeholders

    # Get the placeholders
    directory.placeholders

    # Make a copy of the directory with the placeholders swapped out
    directory.make_copy_with_swaps(
        "path/to/new/directory",
        placeholder="value"
    )
"""

import os

from dir_wand.file import File


class Directory:
    """
    A class for defining the directory.

    A directory is a tree structure containing files and other directories.
    Once instantiated, the directory will unpack its contents and build the
    tree structure.

    A single call to `make_copy_with_swaps` will make a copy of the directory
    and any files it contains including any children directories with the
    placeholders swapped out.

    Attributes:
        path (str):
            The path to the directory.
        children (list):
            A list of child directories.
        files (list):
            A list of files in the directory.
    """

    def __init__(self, path):
        """
        Create the directory instance.

        Args:
            path (str):
                The path to the directory.
        """
        self.path = path
        self.children = []
        self.files = []

    def __str__(self):
        """
        Return the string representation of the directory.

        Returns:
            str:
                The string representation of the directory.
        """
        return self._str_helper()

    def _str_helper(self, prefix="", is_last=True):
        """
        Build the string representation with lines and indentation.

        Args:
            prefix (str):
                The prefix to add to the string.
            is_last (bool):
                Whether the directory is the last in the list.
        """
        connector = "└── " if is_last else "├── "
        result = f"{prefix}{connector}{os.path.basename(self.path)}/\n"
        prefix += "    " if is_last else "│   "

        entries = self.files + self.children
        for index, entry in enumerate(entries):
            is_last_entry = index == len(entries) - 1
            if isinstance(entry, File):
                file_connector = "└── " if is_last_entry else "├── "
                result += f"{prefix}{file_connector}{entry.name}\n"
            else:
                result += entry._str_helper(prefix, is_last_entry)

        return result

    @property
    def name(self):
        """Return the name of the directory."""
        return os.path.basename(self.path)

    @property
    def is_empty(self):
        """Return whether the directory is empty."""
        return not (self.children or self.files)

    @property
    def contains_placeholders(self):
        """Return whether the directory contains placeholders."""
        return any(file.contains_placeholders for file in self.files)

    @property
    def placeholders(self):
        """Return the placeholders in the directory."""
        placeholders = set()

        for file in self.files:
            placeholders.update(file.placeholders)

        return placeholders

    def unpack_contents(self):
        """Unpack the contents of the directory."""
        # Get the contents
        contents = os.listdir(self.path)

        # Iterate over the contents
        for item in contents:
            # Get the full path
            item_path = os.path.join(self.path, item)

            # Check if the item is a directory
            if os.path.isdir(item_path):
                # Create a new directory instance
                directory = Directory(item_path)

                # Add the directory to the children
                self.children.append(directory)

                # Unpack the contents of the directory
                directory.unpack_contents()

            # Check if the item is a file
            elif os.path.isfile(item_path):
                # Create a new file instance
                f = File(item_path)

                # Add the file to the files
                self.files.append(f)

    def make_copy_with_swaps(self, path, **swaps):
        """
        Make a copy of the directory with the placeholders swapped out.

        Args:
            path (str):
                The path to the new directory.
            **swaps:
                The placeholders to swap out.
        """
        # Get the new file path handling any possible placeholders
        path += "/" if not path.endswith("/") else ""
        path = path + self.name.format(**swaps)

        # Make the new directory
        os.makedirs(path, exist_ok=True)

        # Iterate over the files
        for file in self.files:
            # Make a copy of the file with the swaps
            file.make_copy_with_swaps(path, **swaps)

        # Iterate over the children
        for child in self.children:
            # Make a copy of the child with the swaps
            child.make_copy_with_swaps(path, **swaps)



"""The main runner for Directory WAND.

This module contains the main runner for Directory WAND. This is the entry
point for the template copying command line interface.

Example:
    $ python -m dir_wand path/to/template --root path/to/root
    --swap1 value1 --swap2 value2 -swap3 value3 value4 value5
"""

from dir_wand.parser import Parser
from dir_wand.template import Template


def main():
    """Wave the Directory WAND and make magic happen."""
    parser = Parser(
        description="Wave your Directory WAND and make magic happen."
    )

    # Get the arguments
    args = parser.parse_args()

    # Report what we found
    print("Template:", args.template)
    print("Root:", args.root)
    for key, value in args.replacements.items():
        print(f"Replacement: {key} = {value}")

    # Create the template
    template = Template(args.template, **args.replacements)
    print(template)

    # Make the copy
    template.make_copies(args.root)

    # Report we're done
    print("Done!")



"""A module containing the class defining the template.

A Template contains all the information to make copies of a template directory
structure. This is main interface from which the tree of directories and files
will be constructed and can later be copied with the placeholders swapped out.

Example:
    # Create a template instance
    template = Template("path/to/template", swap1=[1, 2, 3], swap2=[4, 5, 7])

    # Make copies of the template
    template.make_copies("path/to/new/template")

"""

import os

from dir_wand.directory import Directory


class Template:
    """
    A class for defining the template.

    A template is the controller class that will manage the directory tree
    making copies of the template with the placeholders swapped out, and
    running any commands that need to be executed.

    Attributes:
        root_path (str):
            The path to the root of the template.
        root (str):
            The name of the root directory.
        directory (Directory):
            The root directory of the template.
        swaps (dict):
            A dictionary of the swaps to make.
        nswap_vars (int):
            The number of swap variables.
        nswaps (int):
            The number of swaps to make.
    """

    def __init__(self, root, **swaps):
        """
        Create the template instance.

        Args:
            root (str):
                The path to the root of the template.
            **swaps (dict):
                The swaps to make in the template.
        """
        # Attach a pointer to the root of the template
        self.root_path = root
        self.root = os.path.basename(root)

        # We will store the directory structure as a tree, by creating
        # the root we'll recursively collect the whole directory structure
        # including files
        self.directory = Directory(self.root_path)
        self.directory.unpack_contents()

        # Print the directory structure
        print("Template structure:")
        print(self.directory)

        # Parse the swaps, we'll store these in a dictionary
        self.swaps = self.parse_swaps(**swaps)
        self.nswap_vars = len(self.swaps)
        self.nswaps = len(list(self.swaps.values())[0])

    def __str__(self):
        """
        Return the string representation of the template.

        Returns:
            str:
                The string representation of the template.
        """
        swaps = ", ".join(
            f"{key}={value}" for key, value in self.swaps.items()
        )
        return (
            f"Template(template_path={self.root_path},"
            f" root={self.root}, swaps={swaps})"
        )

    def parse_swaps(self, **swaps):
        """
        Parse the swaps.

        Args:
            **swaps (dict):
                The swaps to make in the template.

        Raises:
            ValueError:
                If the number of swaps isn't equal between all placeholders.
        """
        for key, value in swaps.items():
            # Do we have a list?
            if isinstance(value, (list, tuple)):
                swaps[key] = value

            # Do we have the defintion of a range, i.e. 1-10?
            elif "-" in value:
                # Split the range
                start, end = value.split("-")

                # Convert the range to a dictionary
                swaps[key] = list(range(int(start), int(end) + 1))
            else:
                raise ValueError(f"Invalid swap value: {value}")

        # Ensure we have the same number of elements for all swaps
        lengths = {len(value) for value in swaps.values()}
        if len(lengths) > 1:
            raise ValueError("All swaps must have the same number of elements")

        return swaps

    def make_copies(self, new_root):
        """
        Make copies of the template.

        This method is the main interface to copying the template. It will call
        the copy method of the template root directory, which will recursively
        copy all the directories and files in the template.

        Args:
            new_root (str):
                The path to the new root directory.
        """
        # Unpack all the swaps into each set of directories we'll need to make
        swaps = [
            {swap: self.swaps[swap][i] for swap in self.swaps}
            for i in range(self.nswaps)
        ]

        # Loop over the swaps we'll have to make
        for swap in swaps:
            # Make a copy of the root directory, this will recursively copy all
            # the files and directories handling the swaps
            self.directory.make_copy_with_swaps(new_root, **swap)
